"""Get data from 3DEP database."""
import tempfile
import uuid
from pathlib import Path
from typing import Dict, List, Tuple, Union

import async_retriever as ar
import cytoolz as tlz
import numpy as np
import pygeoutils as geoutils
import pyproj
import rasterio as rio
import rasterio.warp as rio_warp
import xarray as xr
from pydantic import BaseModel, validator
from pygeoogc import WMS, InvalidInputType, InvalidInputValue, ServiceURL, utils
from shapely.geometry import MultiPolygon, Polygon

try:
    import richdem as rd
except ImportError:
    rd = None

from .exceptions import MissingDependency

DEF_CRS = "epsg:4326"

__all__ = ["get_map", "elevation_bygrid", "elevation_bycoords", "deg2mpm"]


def get_map(
    layers: Union[str, List[str]],
    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
    resolution: float,
    geo_crs: str = DEF_CRS,
    crs: str = DEF_CRS,
) -> Union[xr.DataArray, xr.Dataset]:
    """Access to `3DEP <https://www.usgs.gov/core-science-systems/ngp/3dep>`__ service.

    The 3DEP service has multi-resolution sources so depending on the user
    provided resolution the data is resampled on server-side based
    on all the available data sources. The following layers are available:

    - ``DEM``
    - ``Hillshade Gray``
    - ``Aspect Degrees``
    - ``Aspect Map``
    - ``GreyHillshade_elevationFill``
    - ``Hillshade Multidirectional``
    - ``Slope Map``
    - ``Slope Degrees``
    - ``Hillshade Elevation Tinted``
    - ``Height Ellipsoidal``
    - ``Contour 25``
    - ``Contour Smoothed 25``

    Parameters
    ----------
    layers : str or list
        A valid 3DEP layer or a list of them.
    geometry : Polygon, MultiPolygon, or tuple
        A shapely Polygon or a bounding box ``(west, south, east, north)``.
    resolution : float
        The data resolution in meters. The width and height of the output are computed in
        pixels
        based on the geometry bounds and the given resolution.
    geo_crs : str, optional
        The spatial reference system of the input geometry, defaults to
        ``EPSG:4326``.
    crs : str, optional
        The spatial reference system to be used for requesting the data, defaults to
        ``EPSG:4326``. Valis values are ``epsg:4326``, ``epsg:3576``, ``epsg:3571``,
        ``epsg:3575``, ``epsg:3857``, ``epsg:3572``, ``crs:84``, ``epsg:3573``,
        and ``epsg:3574``.

    Returns
    -------
    dict
        A dict where the keys are the layer name and values are the returned response
        from the WMS service as bytes. You can use ``utils.create_dataset`` function
        to convert the responses to ``xarray.Dataset``.
    """
    _geometry = geoutils.geo2polygon(geometry, geo_crs, crs)

    _layers = layers if isinstance(layers, list) else [layers]
    if "DEM" in _layers:
        _layers[_layers.index("DEM")] = "None"

    _layers = [f"3DEPElevation:{lyr}" for lyr in _layers]

    wms = WMS(ServiceURL().wms.nm_3dep, layers=_layers, outformat="image/tiff", crs=crs)
    r_dict = wms.getmap_bybox(_geometry.bounds, resolution, box_crs=crs)

    ds = geoutils.gtiff2xarray(r_dict, _geometry, crs)

    valid_layers = wms.get_validlayers()
    rename = {lyr: lyr.split(":")[-1].replace(" ", "_").lower() for lyr in valid_layers}
    rename.update({"3DEPElevation:None": "elevation"})

    if isinstance(ds, xr.DataArray):
        ds.name = rename[ds.name]
    else:
        ds = ds.rename({n: rename[n] for n in ds.keys()})

    return ds


def elevation_bygrid(
    xcoords: List[float],
    ycoords: List[float],
    crs: str,
    resolution: float,
    depression_filling: bool = False,
) -> xr.DataArray:
    """Get elevation from DEM data for a grid.

    This function is intended for getting elevations for a gridded dataset.

    Parameters
    ----------
    xcoords : list
        List x-coordinates of a a grid.
    ycoords : list
        List of y-coordinates of a grid.
    crs : str
        The spatial reference system of the input grid, defaults to ``EPSG:4326``.
    resolution : float
        The accuracy of the output, defaults to 10 m which is the highest
        available resolution that covers CONUS. Note that higher resolution
        increases computation time so chose this value with caution.
    depression_filling : bool, optional
        Fill depressions before sampling using
        `RichDEM <https://richdem.readthedocs.io/en/latest/>`__ package, defaults to False.

    Returns
    -------
    xarray.DataArray
        An data array with name elevation and the given dim names.
    """
    bbox = (min(xcoords), min(ycoords), max(xcoords), max(ycoords))

    ratio_min = 0.01
    ratio_x = abs((bbox[2] - bbox[0]) / bbox[0])
    ratio_y = abs((bbox[3] - bbox[1]) / bbox[1])
    if (ratio_x < ratio_min) or (ratio_y < ratio_min):
        rad = ratio_min * abs(bbox[0])
        bbox = (bbox[0] - rad, bbox[1] - rad, bbox[2] + rad, bbox[3] + rad)

    wms = WMS(
        ServiceURL().wms.nm_3dep, layers="3DEPElevation:None", outformat="image/tiff", crs=DEF_CRS
    )
    r_dict = wms.getmap_bybox(bbox, resolution, box_crs=crs)
    dem = _reproject_gtiff(r_dict, crs)

    if depression_filling:
        dem = fill_depressions(dem)

    return dem.interp(x=xcoords, y=ycoords)


def _reproject_gtiff(
    r_dict: Dict[str, bytes],
    crs: str,
) -> np.ndarray:
    """Sample a tiff response for a list of coordinates.

    Parameters
    ----------
    r_dict : dict
        Dictionary of (Geo)Tiff byte responses where keys are some names that are used
        for naming each responses, and values are bytes.
    crs : str
        The target spatial reference system.

    Returns
    -------
    numpy.ndarray
        An array of elevations where its index matches the input coords list.
    """
    tmp_dir = tempfile.gettempdir()
    var_name = {lyr: "_".join(lyr.split("_")[:-3]) for lyr in r_dict.keys()}
    attrs = geoutils.get_gtiff_attrs(next(iter(r_dict.values())), ("y", "x"))

    def to_dataset(lyr: str, resp: bytes) -> xr.DataArray:
        with rio.MemoryFile() as memfile:
            memfile.write(resp)
            with memfile.open() as src:
                transform, width, height = rio_warp.calculate_default_transform(
                    src.crs, crs, src.width, src.height, *src.bounds
                )
                kwargs = src.meta.copy()
                kwargs.update(
                    {"crs": crs, "transform": transform, "width": width, "height": height}
                )

                with rio.vrt.WarpedVRT(src, **kwargs) as vrt:
                    if crs != src.crs:
                        for i in range(1, src.count + 1):
                            rio_warp.reproject(
                                source=rio.band(src, i),
                                destination=rio.band(vrt, i),
                                src_transform=src.transform,
                                src_crs=src.crs,
                                dst_transform=transform,
                                crs=crs,
                                resampling=rio_warp.Resampling.bilinear,
                            )
                    ds = xr.open_rasterio(vrt)
                    ds = ds.squeeze("band", drop=True)
                    ds = ds.sortby("y", ascending=False)
                    ds.attrs["crs"] = crs
                    ds.name = var_name[lyr]
                    fpath = Path(tmp_dir, f"{uuid.uuid4().hex}.nc")
                    ds.to_netcdf(fpath)
                    return fpath

    ds = xr.open_mfdataset(
        (to_dataset(lyr, resp) for lyr, resp in r_dict.items()),
        parallel=True,
    )
    ds = ds[list(ds.keys())[0]]
    ds.name = "elevation"
    ds.attrs["units"] = "meters"
    ds.attrs["crs"] = crs
    ds.attrs["nodatavals"] = (attrs.nodata,)
    transform, _, _ = geoutils.get_transform(ds, attrs.dims)
    ds = ds.sortby(attrs.dims[0], ascending=False)
    ds.attrs["transform"] = transform
    ds.attrs["res"] = (transform.a, transform.e)
    return ds


def fill_depressions(
    dem: Union[xr.DataArray, np.ndarray],
) -> xr.DataArray:
    """Fill depressions and adjust flat areas in a DEM using RichDEM.

    Parameters
    ----------
    dem : xarray.DataArray or numpy.ndarray
        Digital Elevation Model.

    Returns
    -------
    xarray.DataArray
        Conditioned DEM.
    """
    if rd is None:
        raise MissingDependency

    rda = rd.rdarray(dem, no_data=dem.nodatavals[0])
    rda.projection = dem.crs
    rda.geotransform = dem.transform
    rda = rd.FillDepressions(rda, epsilon=False)
    dem.data = rd.ResolveFlats(rda)
    return dem


class ElevationByCoords(BaseModel):
    """Elevation model class.

    Parameters
    ----------
    coords : list of tuple
        List of coordinates.
    crs : str, optional
        Coordinate reference system of the input coordinates, defaults to ``EPSG:4326``.
    source : str, optional
        Elevation source, defaults to ``tnm``. Valid sources are: ``tnm``, ``airmap``.
    """

    crs: str = DEF_CRS
    coords: List[Tuple[float, float]]
    source: str = "airmap"

    @validator("crs")
    def _valid_crs(cls, v):
        try:
            return pyproj.CRS(v)
        except pyproj.exceptions.CRSError as ex:
            raise InvalidInputType("crs", "a valid CRS") from ex

    @validator("coords")
    def _validate_coords(cls, v, values):
        return utils.match_crs(v, values["crs"], DEF_CRS)

    @validator("source")
    def _validate_source(cls, v):
        valid_sources = ["tnm", "airmap"]
        if v not in valid_sources:
            raise InvalidInputValue("source", valid_sources)
        return v

    def airmap(self) -> List[Tuple[float, float]]:
        """Return list of elevations in meters."""
        coords_chunks = tlz.partition_all(100, self.coords)
        headers = {"Content-Type": "application/json", "charset": "utf-8"}
        urls, kwds = zip(
            *(
                (
                    ServiceURL().restful.airmap,
                    {
                        "params": {"points": ",".join(f"{lat},{lon}" for lon, lat in chunk)},
                        "headers": headers,
                    },
                )
                for chunk in coords_chunks
            )
        )
        elevations = list(tlz.pluck("data", ar.retrieve(urls, "json", kwds)))
        return list(tlz.concat(elevations))

    def tnm(self) -> List[Tuple[float, float]]:
        """Return list of elevations in meters."""
        urls, kwds = zip(
            *(
                (
                    ServiceURL().restful.nm_pqs,
                    {
                        "params": {
                            "units": "meters",
                            "output": "json",
                            "y": f"{lat:.5f}",
                            "x": f"{lon:.5f}",
                        },
                    },
                )
                for lon, lat in self.coords
            )
        )
        resp = ar.retrieve(urls, "json", kwds, max_workers=500)
        return [
            r["USGS_Elevation_Point_Query_Service"]["Elevation_Query"]["Elevation"] for r in resp
        ]


def elevation_bycoords(
    coords: List[Tuple[float, float]], crs: str = DEF_CRS, source: str = "tnm"
) -> List[float]:
    """Get elevation from Airmap at 1-arc resolution (~30 m) for a list of coordinates.

    Parameters
    ----------
    coords : list of tuples
        Coordinates of target location as list of tuples ``[(x, y), ...]``.
    crs : str, optional
        Spatial reference (CRS) of coords, defaults to ``EPSG:4326``.
    source : str, optional
        Data source to be used, default to ``tnm``. Supported sources are
        ``airmap`` (30 m resolution) and ``tnm`` (using The National Map's Bulk Point
        Query Service). The ``tnm`` source is more accurate since it uses the highest
        available resolution DEM automatically but it is limited to the US.

    Returns
    -------
    list of float
        Elevation in meter.
    """
    service = ElevationByCoords(crs=crs, coords=coords, source=source)
    return getattr(service, source)()


def deg2mpm(da: xr.DataArray) -> xr.DataArray:
    """Convert slope from degree to meter/meter."""
    attrs = da.attrs
    da = np.tan(np.deg2rad(da))
    da.attrs = attrs
    da.name = "slope"
    da.attrs["units"] = "meters/meters"
    return da



"""Command-line interface for Py3DEP."""
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import click
import geopandas as gpd
import pandas as pd
from shapely.geometry import MultiPolygon, Polygon

from . import py3dep
from .exceptions import MissingColumns, MissingOption


def get_target_df(
    tdf: Union[pd.DataFrame, gpd.GeoDataFrame], req_cols: List[str]
) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
    """Check if all required columns exists in the dataframe.

    It also re-orders the columns based on ``req_cols`` order.
    """
    missing = [c for c in req_cols if c not in tdf]
    if len(missing) > 0:
        raise MissingColumns(missing)
    return tdf[req_cols]


def from_geometry(
    layer: str,
    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
    res: float,
    crs: str,
    nc_path: Union[str, Path],
) -> None:
    """Get topographic data from 3DEP for a geometry."""
    py3dep.get_map(layer, geometry, res, geo_crs=crs, crs=crs).to_netcdf(nc_path)


def from_coords(
    coords: List[Tuple[float, float]], crs: str, query_source: str, csv_path: Union[str, Path]
) -> None:
    """Get elevations of a set of coordinates in meter from airmap."""
    elev = pd.DataFrame.from_records(coords, columns=["x", "y"])
    elev["elevation"] = py3dep.elevation_bycoords(coords, crs, query_source)
    elev.astype("f8").to_csv(csv_path)


LAYERS = [
    "DEM",
    "Hillshade Gray",
    "Aspect Degrees",
    "Aspect Map",
    "GreyHillshade_elevationFill",
    "Hillshade Multidirectional",
    "Slope Map",
    "Slope Degrees",
    "Hillshade Elevation Tinted",
    "Height Ellipsoidal",
    "Contour 25",
    "Contour Smoothed 25",
]

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("target", type=click.Path(exists=True))
@click.argument("target_type", type=click.Choice(["geometry", "coords"], case_sensitive=False))
@click.argument("crs", type=str)
@click.option(
    "-q",
    "--query_source",
    default="airmap",
    type=click.Choice(["airmap", "tnm"], case_sensitive=False),
    help="Source of the elevation data.",
)
@click.option(
    "-l",
    "--layer",
    default=None,
    type=click.Choice(LAYERS, case_sensitive=True),
    help="Layer name when requesting for topographic data.",
)
@click.option(
    "-s",
    "--save_dir",
    type=click.Path(exists=False),
    default="topo_3dep",
    help="Path to a directory to save the requested files. Extension for the outputs is .nc for geometry and .csv for coords.",
)
def main(
    target: Path,
    target_type: str,
    crs: str,
    query_source: str = "airmap",
    layer: Optional[str] = None,
    save_dir: Union[str, Path] = "topo_3dep",
):
    r"""Retrieve topographic data within geometries or elevations for a list of coordinates.

    TARGET: Path to a geospatial file (any file that geopandas.read_file can open) or a csv file.

    The geospatial file should have three columns:

        - id: Feature identifiers that py3dep uses as the output netcdf/csv filenames.

        - res: Target resolution in meters.

        - geometry: A Polygon or MultiPloygon.

    The csv file should have two column: x and y.

    TARGET_TYPE: Type of input file: "coords" for csv and "geometry" for geospatial.

    CRS: CRS of the input data.

    Examples:

        $ py3dep ny_coords.csv coords epsg:4326

        $ py3dep ny_geom.gpkg geometry epsg:3857 --layer "Slope Map"
    """  # noqa: D412
    save_dir = Path(save_dir)
    if not save_dir.exists():
        os.makedirs(save_dir, exist_ok=True)

    target = Path(target)
    if target_type == "geometry":
        if layer is None:
            raise MissingOption

        target_df = gpd.read_file(target, crs=crs)
        target_df = get_target_df(target_df, ["id", "res", "geometry"])
        args_list = (
            (layer, g, r, crs, Path(save_dir, f"{i}.nc"))
            for i, r, g in target_df.itertuples(index=False)
        )

        click.echo(f"Found {len(target_df)} items in {target}. Retrieving ...")
        with click.progressbar(
            args_list, label=f"Getting {layer} from 3DEP", length=len(target_df)
        ) as bar:
            for args in bar:
                from_geometry(*args)

        click.echo(f"Retrieved topography data for {len(target_df)} item(s).")
    else:
        target_df = pd.read_csv(target)
        target_df = get_target_df(target_df, ["x", "y"])
        click.echo(f"Found {len(target_df)} items in {target}. Retrieving ...")
        from_coords(
            list(target_df.itertuples(index=False, name=None)),
            crs,
            query_source,
            Path(save_dir, f"{target.stem}_elevation.csv"),
        )
        click.echo(f"Retrieved elevation data for {len(target_df)} item(s).")



"""Customized Py3DEP exceptions."""
from typing import List


class MissingColumns(Exception):
    """Exception raised when a required column is missing from a dataframe.

    Parameters
    ----------
    missing : list
        List of missing columns.
    """

    def __init__(self, missing: List[str]) -> None:
        self.message = "The following columns are missing:\n" + f"{', '.join(missing)}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class MissingDependency(ImportError):
    """Exception raised when RichDEM is not installed."""

    def __init__(self) -> None:
        self.message = "Depression filling operation uses richdem package which is not installed."
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class MissingOption(Exception):
    """Exception raised when layer is not provided."""

    def __init__(self) -> None:
        self.message = "layer option is required when target_type is geometry."
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



import nox


@nox.session(python="3.9")
def tests(session):
    session.install(".[test,dem]")
    hr_deps = ["pygeoogc", "pygeoutils"]
    for p in hr_deps:
        session.install(f"git+https://github.com/cheginit/{p}.git")
    session.run("pytest")
    session.run("coverage", "report")
    session.run("coverage", "html")



import io
import shutil
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
from pygeoogc import utils
from shapely.geometry import Polygon

import py3dep

DEF_CRS = "epsg:4326"
ALT_CRS = "epsg:3857"
GEOM = Polygon(
    [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45], [-69.77, 45.45], [-69.77, 45.07]]
)
LYR = "Slope Degrees"
SMALL = 1e-3


def test_getmap():
    layers = ["DEM", LYR]
    ds = py3dep.get_map(layers, GEOM.bounds, 1e3, geo_crs=DEF_CRS, crs=ALT_CRS)
    dem_10 = py3dep.get_map(layers[0], GEOM, 10, geo_crs=DEF_CRS, crs=ALT_CRS)
    dem_1e3 = py3dep.get_map(layers[0], GEOM, 1e3, geo_crs=DEF_CRS, crs=ALT_CRS)
    assert (
        sorted(ds.keys()) == ["elevation", "slope_degrees"]
        and abs(dem_10.mean().compute().item() - dem_1e3.mean().compute().item()) < 7e-2
    )


class TestByCoords:
    coords = [(-7766049.664788851, 5691929.739021257)] * 200

    def test_airmap(self):
        airmap = py3dep.elevation_bycoords(self.coords, ALT_CRS, source="airmap")
        assert set(airmap) == {363}

    def test_tnm(self):
        tnm = py3dep.elevation_bycoords(self.coords, ALT_CRS, source="tnm")
        assert set(tnm) == {1169.9}


def test_deg2mpm():
    slope = py3dep.get_map(LYR, GEOM, 1e3)
    slope = py3dep.deg2mpm(slope)
    assert abs(slope.mean().compute().item() - 0.05) < SMALL


def test_grid():
    crs = "+proj=lcc +lat_1=25 +lat_2=60 +lat_0=42.5 +lon_0=-100 +x_0=0 +y_0=0 +ellps=WGS84 +units=m +no_defs"
    geom = utils.match_crs(GEOM, DEF_CRS, crs)
    xmin, ymin, xmax, ymax = geom.bounds
    res = 1e3
    gx = np.arange(xmin, xmax, res)
    gy = np.arange(ymin, ymax, res)
    elev = py3dep.elevation_bygrid(gx, gy, crs, res)
    elev_fill = py3dep.elevation_bygrid(gx, gy, crs, res, depression_filling=True)
    assert ((elev_fill - elev).sum().compute().item() - 1935.074) < SMALL


def test_cli_map(script_runner):
    gdf = gpd.GeoDataFrame({"id": "geo_test", "res": 1e3}, geometry=[GEOM], index=[0])
    geo_gpkg = "nat_geo.gpkg"
    gdf.to_file(geo_gpkg)
    ret = script_runner.run("py3dep", geo_gpkg, "geometry", DEF_CRS, "-l", LYR, "-s", "geo_map")
    shutil.rmtree(geo_gpkg)
    shutil.rmtree("geo_map")
    assert ret.success
    assert "Retrieved topography data for 1 item(s)." in ret.stdout
    assert ret.stderr == ""


def test_cli_coords(script_runner):
    df = pd.DataFrame([(-7766049.664788851, 5691929.739021257)] * 3, columns=["x", "y"])
    coord_csv = "coords.csv"
    df.to_csv(coord_csv)
    ret = script_runner.run("py3dep", coord_csv, "coords", ALT_CRS, "-s", "geo_coords")
    Path(coord_csv).unlink()
    shutil.rmtree("geo_coords")
    assert ret.success
    assert "Retrieved elevation data for 3 item(s)." in ret.stdout
    assert ret.stderr == ""


def test_show_versions():
    f = io.StringIO()
    py3dep.show_versions(file=f)
    assert "INSTALLED VERSIONS" in f.getvalue()



"""Top-level package for Py3DEP."""
from pkg_resources import DistributionNotFound, get_distribution

from .exceptions import MissingColumns, MissingDependency, MissingOption
from .print_versions import show_versions
from .py3dep import deg2mpm, elevation_bycoords, elevation_bygrid, get_map

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = "999"

__all__ = [
    # Functions
    "get_map",
    "deg2mpm",
    "elevation_bycoords",
    "elevation_bygrid",
    "show_versions",
    # Exceptions
    "MissingColumns",
    "MissingOption",
    "MissingDependency",
    # Constants
    "__version__",
]
